# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
import sys
from ast import (
    alias,
    AnnAssign,
    arg,
    Assign,
    AST,
    AsyncFunctionDef,
    Attribute,
    Call,
    ClassDef,
    Constant,
    copy_location,
    DictComp,
    expr,
    FunctionDef,
    GeneratorExp,
    Global,
    Import,
    ImportFrom,
    Lambda,
    ListComp,
    Module,
    Name,
    NodeVisitor,
    SetComp,
    stmt,
    Try,
)
from symtable import SymbolTable
from types import ModuleType
from typing import (
    cast,
    final,
    Generic,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from ..common import (
    AstRewriter,
    get_symbol_map,
    imported_name,
    lineinfo,
    mangle_priv_name,
    ScopeStack,
    SymbolMap,
    SymbolScope,
)


# We want to transform a module which looks something like:
#
# from strict_modules import cached_property
#
# x = 1
# def f():
#     return min(x, 20)
#
# class C:
#     @cached_property
#     def f(self): return 42
#
# C.foo = 42
#
# def g():
#     class C: pass
#     return C
#
# Into something like:
#
# min = <builtins>.min  # we have a fix set of builtins we execute against
#
# # We want some helper methods to be available to us:
# <strict-modules> = <fixed-modules>["strict_modules"]
#
# x = 1
# def f():
#     return min(x, 20)
#
# @<init-cached-properties>({'f': '_f_impl'})
# class C:
#     __slots__ = ('f', )

#     def _f_impl(self):
#         return 42
#
#
# C.foo = 42
#
# def g():
#     class C: pass
#     return C
#     <classes> = []
#
#
#
# # Top-level assigned names
#
# Note: we use non-valid Python identifeirs for anything which we introduce like
# <strict-module> and use valid Python identifiers for variables which will be accessible
# from within the module.


def make_arg(name: str) -> arg:
    return lineinfo(ast.arg(name, None))


TAst = TypeVar("TAst", bound=AST)


def make_assign(*a: object, **kw: object) -> Assign:
    # pyre-fixme[6]: For 1st argument expected `List[expr]` but got `object`.
    # pyre-fixme[6]: For 2nd argument expected `expr` but got `object`.
    # pyre-fixme[6]: For 3rd argument expected `Optional[str]` but got `object`.
    node = Assign(*a, **kw)
    node.type_comment = None
    return node


def copyline(from_node: AST, to_node: TAst) -> TAst:
    # pyre-fixme[16]: `AST` has no attribute `lineno`.
    to_node.lineno = from_node.lineno
    # pyre-fixme[16]: `AST` has no attribute `col_offset`.
    to_node.col_offset = from_node.col_offset
    return to_node


_IMPLICIT_GLOBALS = [
    "__name__",
    "__loader__",
    "__package__",
    "__spec__",
    "__path__",
    "__file__",
    "__cached__",
]


def make_function(name: str, pos_args: list[arg]) -> FunctionDef:
    # pyre-fixme[20]: Argument `args` expected.
    func = lineinfo(ast.FunctionDef())
    func.name = name
    # pyre-fixme[20]: Argument `args` expected.
    args = ast.arguments()
    args.kwonlyargs = []
    args.kw_defaults = []
    args.defaults = []
    args.args = pos_args
    args.posonlyargs = []
    func.args = args
    args.kwarg = None
    args.vararg = None
    func.decorator_list = []
    func.returns = None
    func.type_comment = ""
    return func


def make_assign_empty_list(name: str) -> Assign:
    return lineinfo(
        make_assign(
            [lineinfo(ast.Name(name, ast.Store()))],
            lineinfo(ast.List([], ast.Load())),
        )
    )


GLOBALS_HELPER_ALIAS = "<globals-helper>"


@final
class StrictModuleRewriter:
    """rewrites a module body so that all global variables are transformed into
    local variables, and are closed over by the enclosing functions.  This will
    ultimately remove all LOAD_GLOBAL/STORE_GLOBAL opcodes, and therefore will
    also have the side effect of making the module read-only as globals will
    not be exposed."""

    def __init__(
        self,
        root: Module,
        table: SymbolTable,
        filename: str,
        modname: str,
        mode: str,
        optimize: int,
        builtins: ModuleType | Mapping[str, object] = __builtins__,
        is_static: bool = False,
    ) -> None:
        if not isinstance(builtins, dict):
            builtins = builtins.__dict__

        self.root = root
        self.table = table
        self.filename = filename
        self.modname = modname
        self.mode = mode
        self.optimize = optimize
        self.builtins: Mapping[str, object] = builtins
        self.symbol_map: SymbolMap = get_symbol_map(root, table)
        scope: SymbolScope[None, None] = SymbolScope(table, None)
        self.visitor = ImmutableVisitor(ScopeStack(scope, symbol_map=self.symbol_map))
        # Top-level statements in the returned code object...
        self.code_stmts: list[stmt] = []
        self.is_static = is_static

    def transform(self) -> ast.Module:
        original_first_node = self.root.body[0] if self.root.body else None
        self.visitor.visit(self.root)
        self.visitor.global_sets.update(_IMPLICIT_GLOBALS)

        for argname in _IMPLICIT_GLOBALS:
            self.visitor.globals.add(argname)

        # pyre-fixme[20]: Argument `type_ignores` expected.
        mod = ast.Module(
            [
                *self.get_future_imports(),
                *self.transform_body(),
            ]
        )
        if mod.body and original_first_node:
            # this isn't obvious but the new mod body is empty
            # if the original module body is empty. Therefore there
            # is always a location to copy
            copy_location(mod.body[0], original_first_node)

        mod.type_ignores = []
        return mod

    def get_future_imports(self) -> Iterable[stmt]:
        if self.visitor.future_imports:
            yield lineinfo(
                ImportFrom("__future__", list(self.visitor.future_imports), 0)
            )

    def del_global(self, name: str) -> stmt:
        return lineinfo(ast.Delete([lineinfo(ast.Name(name, ast.Del()))]))

    def store_global(self, name: str, value: expr) -> stmt:
        return lineinfo(make_assign([lineinfo(ast.Name(name, ast.Store()))], value))

    def load_global(self, name: str) -> expr:
        return lineinfo(ast.Name(name, ast.Load()))

    def create_annotations(self) -> stmt:
        return self.store_global("__annotations__", lineinfo(ast.Dict([], [])))

    def make_transformer(
        self,
        scopes: ScopeStack[None, ScopeData],
    ) -> ImmutableTransformer:
        return ImmutableTransformer(
            scopes,
            self.modname,
            self.builtins,
            self.visitor.globals,
            self.visitor.global_sets,
            self.visitor.global_dels,
            self.visitor.future_imports,
            self.is_static,
        )

    def transform_body(self) -> Iterable[stmt]:
        scopes = ScopeStack(
            SymbolScope(self.table, ScopeData()), symbol_map=self.symbol_map
        )
        transformer = self.make_transformer(scopes)
        body = transformer.visit(self.root).body

        return body


def rewrite(
    root: Module,
    table: SymbolTable,
    filename: str,
    modname: str,
    mode: str = "exec",
    optimize: int = -1,
    builtins: ModuleType | Mapping[str, object] = __builtins__,
    is_static: bool = False,
) -> Module:
    return StrictModuleRewriter(
        root,
        table,
        filename,
        modname,
        mode,
        optimize,
        builtins,
        is_static=is_static,
    ).transform()


TTransformedStmt = Union[Optional[AST], List[AST]]
TVar = TypeVar("TScope")
TScopeData = TypeVar("TData")


class SymbolVisitor(Generic[TVar, TScopeData], NodeVisitor):
    def __init__(self, scopes: ScopeStack[TVar, TScopeData]) -> None:
        self.scopes = scopes

    @property
    def skip_annotations(self) -> bool:
        return False

    def is_global(self, name: str) -> bool:
        return self.scopes.is_global(name)

    def scope_for(self, name: str) -> SymbolScope[TVar, TScopeData]:
        return self.scopes.scope_for(name)

    def visit_Comp_Outer(
        self,
        node: ListComp | SetComp | GeneratorExp | DictComp,
        update: bool = False,
    ) -> None:
        iter = self.visit(node.generators[0].iter)
        if update:
            node.generators[0].iter = iter

    def visit_Try(self, node: Try) -> TTransformedStmt:
        # Need to match the order the symbol visitor constructs these in, which
        # walks orelse before handlers.
        for val in node.body:
            self.visit(val)

        for val in node.orelse:
            self.visit(val)

        for val in node.handlers:
            self.visit(val)

        for val in node.finalbody:
            self.visit(val)

        return node

    def visit_Comp_Inner(
        self,
        node: ListComp | SetComp | GeneratorExp | DictComp,
        update: bool = False,
        scope_node: ListComp | SetComp | GeneratorExp | DictComp | None = None,
    ) -> None:
        scope_node = scope_node or node
        with self.scopes.with_node_scope(scope_node):
            if isinstance(node, DictComp):
                key = self.visit(node.key)
                if update:
                    node.key = key
                value = self.visit(node.value)
                if update:
                    node.value = value
            else:
                elt = self.visit(node.elt)
                if update:
                    node.elt = elt

            self.walk_many(node.generators[0].ifs, update)

            target = self.visit(node.generators[0].target)
            if update:
                # pyre-fixme[16]: `DictComp` has no attribute `target`.
                # pyre-fixme[16]: `GeneratorExp` has no attribute `target`.
                # pyre-fixme[16]: `ListComp` has no attribute `target`.
                # pyre-fixme[16]: `SetComp` has no attribute `target`.
                node.target = target

            gens = node.generators[1:]
            self.walk_many(gens, update)
            if update:
                node.generators[1:] = gens

    def visit_Func_Outer(
        self, node: AsyncFunctionDef | FunctionDef | Lambda, update: bool = False
    ) -> None:
        args = self.visit(node.args)
        if update:
            node.args = args

        if isinstance(node, (AsyncFunctionDef, FunctionDef)):
            retnode = node.returns
            if retnode and not self.skip_annotations:
                returns = self.visit(retnode)
                if update:
                    node.returns = returns

            self.walk_many(node.decorator_list, update)

    def visit_Func_Inner(
        self,
        node: AsyncFunctionDef | FunctionDef | Lambda,
        update: bool = False,
        scope_node: AsyncFunctionDef | FunctionDef | Lambda | None = None,
    ) -> SymbolScope[TVar, TScopeData]:
        scope_node = scope_node or node
        with self.scopes.with_node_scope(scope_node) as next:
            assert next
            assert isinstance(node, Lambda) or node.name == next.symbols.get_name()

            # visit body in function scope
            if isinstance(node, Lambda):
                new = self.visit(node.body)
                if update:
                    node.body = new
            else:
                self.walk_many(node.body, update)

            return next

    def visit_Class_Outer(self, node: ClassDef, update: bool = False) -> None:
        self.walk_many(node.bases, update)
        self.walk_many(node.keywords, update)
        self.walk_many(node.decorator_list, update)

    def visit_Class_Inner(
        self,
        node: ClassDef,
        update: bool = False,
        scope_node: ClassDef | None = None,
    ) -> SymbolScope[TVar, TScopeData]:
        scope_node = scope_node or node
        with self.scopes.with_node_scope(scope_node) as next:
            assert next
            assert node.name == next.symbols.get_name()
            # visit body in class scope
            self.walk_many(node.body, update)

            return next

    def walk_many(
        self,
        stmts: list[TAst],
        update: bool = False,
    ) -> None:
        new_stmts = []
        for statement in stmts:
            new = self.visit(statement)
            if isinstance(new, ast.AST):
                new_stmts.append(new)
            elif new is not None:
                new_stmts.extend(new)
        if update:
            stmts[:] = new_stmts

    def visit_Lambda(self, node: Lambda) -> TTransformedStmt:
        self.visit_Func_Outer(node)

        self.visit_Func_Inner(node)

    def visit_comp(
        self, node: ListComp | SetComp | DictComp | GeneratorExp
    ) -> expr | None:
        self.visit_Comp_Outer(node)

        self.visit_Comp_Inner(node)

        return node

    def visit_ListComp(self, node: ListComp) -> expr | None:
        return self.visit_comp(node)

    def visit_SetComp(self, node: SetComp) -> expr | None:
        return self.visit_comp(node)

    def visit_GeneratorExp(self, node: GeneratorExp) -> expr | None:
        return self.visit_comp(node)

    def visit_DictComp(self, node: DictComp) -> expr | None:
        return self.visit_comp(node)


@final
class ImmutableVisitor(SymbolVisitor[None, None]):
    def __init__(self, scopes: ScopeStack[None, None]) -> None:
        super().__init__(scopes)
        self.globals: set[str] = set()
        self.global_sets: set[str] = set()
        self.global_dels: set[str] = set()
        self.future_imports: set[alias] = set()

    def future_annotations(self) -> bool:
        for a in self.future_imports:
            if a.name == "annotations":
                return True
        return False

    @property
    def skip_annotations(self) -> bool:
        return self.future_annotations() or sys.version_info >= (3, 14)

    def load_name(self, name: str) -> None:
        if self.is_global(name):
            self.globals.add(name)

    def store_name(self, name: str) -> None:
        if self.is_global(name):
            self.globals.add(name)
            self.global_sets.add(name)

    def del_name(self, name: str) -> None:
        if self.is_global(name):
            self.globals.add(name)
            self.global_sets.add(name)
            self.global_dels.add(name)

    def visit_Global(self, node: Global) -> None:
        for name in node.names:
            self.globals.add(name)

    def visit_Name(self, node: Name) -> None:
        if isinstance(node.ctx, ast.Load):
            self.load_name(node.id)
        elif isinstance(node.ctx, ast.Store):
            self.store_name(node.id)
        elif isinstance(node.ctx, ast.Del):
            self.del_name(node.id)

    def visit_ImportFrom(self, node: ImportFrom) -> None:
        if node.module == "__future__":
            self.future_imports.update(node.names)

        for name in node.names:
            self.store_name(name.asname or name.name)

    def visit_Import(self, node: Import) -> None:
        for name in node.names:
            self.store_name(imported_name(name))

    def visit_Call(self, node: Call) -> None:
        func = node.func
        if isinstance(func, ast.Name):
            # We don't currently allow aliasing or shadowing exec/eval
            # so this check is currently sufficient.
            if (func.id == "exec" or func.id == "eval") and len(node.args) < 2:
                # We'll need access to our globals() helper when we transform
                # the ast
                self.globals.add("globals")
                self.globals.add("locals")
        self.generic_visit(node)

    def visit_Try(self, node: Try) -> None:
        super().visit_Try(node)

        for handler in node.handlers:
            name = handler.name
            if name is None:
                continue
            self.del_name(name)

    def visit_ClassDef(self, node: ClassDef) -> None:
        self.visit_Class_Outer(node)

        self.store_name(node.name)

        self.visit_Class_Inner(node)

    def visit_FunctionDef(self, node: FunctionDef) -> None:
        self.visit_Func_Outer(node)

        self.store_name(node.name)

        self.visit_Func_Inner(node)

    def visit_arg(self, node: ast.arg) -> None:
        if not self.skip_annotations:
            return self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node: AsyncFunctionDef) -> None:
        self.visit_Func_Outer(node)

        self.store_name(node.name)

        self.visit_Func_Inner(node)

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        self.visit(node.target)
        if not self.skip_annotations:
            self.visit(node.annotation)
        value = node.value
        if value:
            self.visit(value)


class ScopeData:
    def __init__(self) -> None:
        self.has_classdefs = False

    def visit_Assign(self, node: Assign) -> None:
        pass

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        pass

    def visit_decorators(self, node: ClassDef | FunctionDef | AsyncFunctionDef) -> None:
        # visit the current node held by the scope data
        pass


@final
class ClassScope(ScopeData):
    def __init__(self, node: ClassDef) -> None:
        super().__init__()
        self.node = node
        self.instance_fields: set[str] = set()
        self.cached_props: dict[str, object] = {}
        self.slots_enabled: bool = False
        self.loose_slots: bool = False
        self.extra_slots: list[str] | None = None

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        target = node.target
        if node.value is None and isinstance(target, Name):
            self.instance_fields.add(target.id)


@final
class FunctionScope(ScopeData):
    def __init__(self, node: AsyncFunctionDef | FunctionDef, parent: ScopeData) -> None:
        super().__init__()
        self.node = node
        self.parent = parent
        self.is_cached_prop: bool = False
        self.cached_prop_value: object = None

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        parent = self.parent
        target = node.target
        if (
            self.node.name == "__init__"
            and self.node.args.args
            and isinstance(parent, ClassScope)
            and isinstance(target, Attribute)
        ):
            self.add_attr_name(target, parent)

    def add_attr_name(self, target: Attribute, parent: ClassScope) -> None:
        """records self.name = ... when salf matches the 1st parameter"""
        value = target.value
        node = self.node
        if isinstance(value, Name) and value.id == node.args.args[0].arg:
            parent.instance_fields.add(target.attr)

    def assign_worker(self, targets: Sequence[AST], parent: ClassScope) -> None:
        for target in targets:
            if isinstance(target, Attribute):
                self.add_attr_name(target, parent)
            elif isinstance(target, (ast.Tuple, ast.List)):
                self.assign_worker(target.elts, parent)

    def visit_Assign(self, node: Assign) -> None:
        parent = self.parent
        if (
            self.node.name == "__init__"
            and isinstance(parent, ClassScope)
            and self.node.args.args
        ):
            self.assign_worker(node.targets, parent)


@final
class ImmutableTransformer(SymbolVisitor[None, ScopeData], AstRewriter):
    def __init__(
        self,
        symbols: ScopeStack[None, ScopeData],
        modname: str,
        builtins: Mapping[str, object],
        globals: set[str],
        global_sets: set[str],
        global_dels: set[str],
        future_imports: set[alias],
        is_static: bool,
    ) -> None:
        super().__init__(symbols)
        symbols.scope_factory = self.make_scope
        self.modname = modname
        self.future_imports = future_imports
        self.is_static = is_static

    def make_scope(
        self,
        symtable: SymbolTable,
        node: AST,
        vars: MutableMapping[str, None] | None = None,
    ) -> SymbolScope[None, ScopeData]:
        if isinstance(node, (FunctionDef, AsyncFunctionDef)):
            data: ScopeData = FunctionScope(node, self.scopes.scopes[-1].scope_data)
        elif isinstance(node, ClassDef):
            data: ScopeData = ClassScope(node)
        else:
            data: ScopeData = ScopeData()
        return SymbolScope(symtable, data)

    def visit_Assign(self, node: Assign) -> TTransformedStmt:
        self.scopes.scopes[-1].scope_data.visit_Assign(node)
        node = self.update_node(
            node,
            targets=self.walk_list(node.targets),
            value=self.visit(node.value),
        )
        return node

    def visit_AnnAssign(self, node: AnnAssign) -> TTransformedStmt:
        self.scopes.scopes[-1].scope_data.visit_AnnAssign(node)
        return self.generic_visit(node)

    def make_base_class_dict_test_stmts(
        self, bases: list[TAst], instance_fields: set[str], location_node: ast.ClassDef
    ) -> stmt:
        slots_stmt = self.make_slots_stmt(instance_fields)
        # if there are non-names in the bases of the class, give up and just create slots
        if not all(isinstance(b, ast.Name) for b in bases):
            return slots_stmt
        # if __dict__ is not added to instance fields (no loose slots), just slotify
        if "__dict__" not in instance_fields:
            return slots_stmt
        # generate code that decide whether __dict__ should be included
        # if any('__dict__' in getattr(_t, '__dict__', ()) for b in <bases> for _t in b.mro()):
        #   __slots__ = <slots without __dict__>
        # else:
        #   __slots__ = <slots with __dict__>
        names = [lineinfo(ast.Name(cast(ast.Name, n).id, ast.Load())) for n in bases]
        condition = lineinfo(
            ast.Call(
                lineinfo(ast.Name("any", ast.Load())),
                [
                    lineinfo(
                        ast.GeneratorExp(
                            lineinfo(
                                ast.Compare(
                                    lineinfo(Constant("__dict__")),
                                    [lineinfo(ast.In())],
                                    [
                                        lineinfo(
                                            ast.Call(
                                                lineinfo(
                                                    ast.Name("getattr", ast.Load())
                                                ),
                                                [
                                                    lineinfo(
                                                        ast.Name("_t", ast.Load())
                                                    ),
                                                    lineinfo(Constant("__dict__")),
                                                    lineinfo(ast.Tuple([], ast.Load())),
                                                ],
                                                [],
                                            )
                                        )
                                    ],
                                )
                            ),
                            [
                                lineinfo(
                                    ast.comprehension(
                                        lineinfo(ast.Name("b", ast.Store())),
                                        # pyre-fixme[6]: For 1st argument expected
                                        #  `List[expr]` but got `List[Name]`.
                                        lineinfo(ast.List(names, ast.Load())),
                                        [],
                                        0,
                                    )
                                ),
                                lineinfo(
                                    ast.comprehension(
                                        lineinfo(ast.Name("_t", ast.Store())),
                                        lineinfo(
                                            ast.Call(
                                                lineinfo(
                                                    ast.Attribute(
                                                        lineinfo(
                                                            ast.Name("b", ast.Load())
                                                        ),
                                                        "mro",
                                                        ast.Load(),
                                                    )
                                                ),
                                                [],
                                                [],
                                            )
                                        ),
                                        [],
                                        0,
                                    )
                                ),
                            ],
                        ),
                        target=location_node,
                    )
                ],
                [],
            )
        )
        slots_stmt_without_dict = self.make_slots_stmt(instance_fields - {"__dict__"})
        return lineinfo(ast.If(condition, [slots_stmt_without_dict], [slots_stmt]))

    def make_slots_stmt(self, instance_fields: set[str]) -> Assign:
        return lineinfo(
            make_assign(
                [lineinfo(Name("__slots__", ast.Store()))],
                lineinfo(
                    ast.Tuple(
                        [lineinfo(Constant(name)) for name in instance_fields],
                        ast.Load(),
                    )
                ),
            )
        )

    def visit_ClassDef(self, node: ClassDef) -> TTransformedStmt:
        outer_scope_data = self.scopes.scopes[-1].scope_data
        outer_scope_data.has_classdefs = True
        orig_node = node
        node = self.clone_node(node)
        self.visit_Class_Outer(node, True)
        class_scope = self.visit_Class_Inner(node, True, scope_node=orig_node)
        scope_data = class_scope.scope_data
        assert isinstance(scope_data, ClassScope), type(class_scope).__name__
        scope_data.visit_decorators(node)

        slots_enabled = scope_data.slots_enabled and not self.is_static
        if slots_enabled:
            if scope_data.loose_slots:
                scope_data.instance_fields.add("__dict__")
                scope_data.instance_fields.add("__loose_slots__")
            extra_slots = scope_data.extra_slots
            if extra_slots:
                scope_data.instance_fields.update(extra_slots)

            node.body.append(
                self.make_base_class_dict_test_stmts(
                    node.bases, scope_data.instance_fields, node
                )
            )

        if scope_data.cached_props and slots_enabled:
            # Add a decorator which replaces our name-mangled cache properties
            # with a class-level decorator that converts them.  We apply it as
            # a decorator so that we get to initialize these before other
            # decorators see the class.
            node.decorator_list.append(
                self.make_cached_property_init_decorator(scope_data, class_scope)
            )

        return node

    def make_cached_property_init_decorator(
        self,
        scope_data: ClassScope,
        class_scope: SymbolScope[TVar, TScopeData],
    ) -> expr:
        return lineinfo(
            ast.Call(
                lineinfo(ast.Name("<init-cached-properties>", ast.Load())),
                [
                    lineinfo(
                        ast.Dict(
                            [
                                lineinfo(
                                    Constant(mangle_priv_name(name, [class_scope]))
                                )
                                for name in scope_data.cached_props
                            ],
                            [
                                lineinfo(
                                    ast.Tuple(
                                        [
                                            lineinfo(
                                                Constant(
                                                    mangle_priv_name(
                                                        self.mangle_cached_prop(name),
                                                        [class_scope],
                                                    )
                                                )
                                            ),
                                            lineinfo(Constant(value)),
                                        ],
                                        ast.Load(),
                                    )
                                )
                                for name, value in scope_data.cached_props.items()
                            ],
                        )
                    )
                ],
                [],
            )
        )

    def mangle_cached_prop(self, name: str) -> str:
        return "_" + name + "_impl"

    def visit_FunctionDef(self, node: FunctionDef) -> TTransformedStmt:
        outer_scope = self.scopes.scopes[-1].scope_data
        orig_node = node
        node = self.clone_node(node)
        self.visit_Func_Outer(node, True)

        scope_data = self.visit_Func_Inner(node, True, scope_node=orig_node).scope_data
        scope_data.visit_decorators(node)

        res: TTransformedStmt = node

        self.check_cached_prop(node, scope_data, outer_scope)

        return res

    def visit_AsyncFunctionDef(self, node: AsyncFunctionDef) -> TTransformedStmt:
        outer_scope = self.scopes.scopes[-1].scope_data
        orig_node = node
        node = self.clone_node(node)
        self.visit_Func_Outer(node, True)

        scope_data = self.visit_Func_Inner(node, True, scope_node=orig_node).scope_data
        scope_data.visit_decorators(node)

        self.check_cached_prop(node, scope_data, outer_scope)

        return node

    def check_cached_prop(
        self,
        node: AsyncFunctionDef | FunctionDef,
        func_scope: ScopeData,
        outer_scope: ScopeData,
    ) -> None:
        if isinstance(func_scope, FunctionScope) and isinstance(
            outer_scope, ClassScope
        ):
            if func_scope.is_cached_prop:
                # preprocessor already checked that outer scope is slotified
                outer_scope.instance_fields.add(node.name)
                outer_scope.cached_props[node.name] = func_scope.cached_prop_value
                node.name = self.mangle_cached_prop(node.name)

    def visit_Lambda(self, node: Lambda) -> TTransformedStmt:
        orig_node = node
        node = self.clone_node(node)
        self.visit_Func_Outer(node, True)

        self.visit_Func_Inner(node, True, scope_node=orig_node)

        return node

    def visit_ImportFrom(self, node: ImportFrom) -> TTransformedStmt:
        if node.module == "__future__":
            # We push these to the top of the module where they're required to be
            return None
        return node

    def visit_comp(
        self, node: ListComp | SetComp | DictComp | GeneratorExp
    ) -> expr | None:
        orig_node = node
        node = self.clone_node(node)
        self.visit_Comp_Outer(node, True)

        self.visit_Comp_Inner(node, True, scope_node=orig_node)

        return node



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# pyre-strict

from contextlib import contextmanager
from typing import Any, AsyncGenerator, Callable, Coroutine, Generator, TypeVar
from warnings import catch_warnings, simplefilter, warn


# The JIT compiles arbitrary Python functions.  Ideally this type would exclude native
# functions, but that doesn't seem possible yet.
#
# pyre-ignore[33]: Not going to add a new type variable for every use of FuncAny.
FuncAny = Callable[..., Any]


try:
    from cinderjit import (
        _deopt_gen,
        append_jit_list,
        auto,
        clear_runtime_stats,
        compile_after_n_calls,
        count_interpreted_calls,
        disable,
        disable_emit_type_annotation_guards,
        disable_hir_inliner,
        disable_specialized_opcodes,
        enable,
        enable_emit_type_annotation_guards,
        enable_hir_inliner,
        enable_specialized_opcodes,
        force_compile,
        force_uncompile,
        get_allocator_stats,
        get_and_clear_inline_cache_stats,
        get_and_clear_runtime_stats,
        get_compilation_time,
        get_compile_after_n_calls,
        get_compiled_functions,
        get_compiled_size,
        get_compiled_spill_stack_size,
        get_compiled_stack_size,
        get_function_compilation_time,
        get_function_hir_opcode_counts,
        get_inlined_functions_stats,
        get_jit_list,
        get_num_inlined_functions,
        is_enabled,
        is_hir_inliner_enabled,
        is_inline_cache_stats_collection_enabled,
        is_jit_compiled,
        jit_suppress,
        jit_unsuppress,
        lazy_compile,
        mlock_profiler_dependencies,
        multithreaded_compile_test,
        page_in_profiler_dependencies,
        precompile_all,
        read_jit_list,
        set_max_code_size,
    )

except ImportError:
    TDeoptGenYield = TypeVar("TDeoptGenYield")
    TDeoptGenSend = TypeVar("TDeoptGenSend")
    TDeoptGenReturn = TypeVar("TDeoptGenReturn")

    def _deopt_gen(
        gen: Generator[TDeoptGenYield, TDeoptGenSend, TDeoptGenReturn]
        | AsyncGenerator[TDeoptGenYield, TDeoptGenSend]
        | Coroutine[TDeoptGenYield, TDeoptGenSend, TDeoptGenReturn],
    ) -> bool:
        return False

    def append_jit_list(entry: str) -> None:
        return None

    def auto() -> None:
        return None

    def clear_runtime_stats() -> None:
        return None

    def compile_after_n_calls(calls: int) -> None:
        return None

    def count_interpreted_calls(func: FuncAny) -> int:
        return 0

    def disable(deopt_all: bool = False) -> None:
        return None

    def disable_emit_type_annotation_guards() -> None:
        return None

    def disable_hir_inliner() -> None:
        return None

    def disable_specialized_opcodes() -> None:
        return None

    def enable() -> None:
        # Warn here because users might think this is function is how to enable the JIT
        # when it is not installed.
        warn(
            "Cinder JIT is not installed, calling cinderx.jit.enable() is doing nothing"
        )

    def enable_emit_type_annotation_guards() -> None:
        return None

    def enable_hir_inliner() -> None:
        return None

    def enable_specialized_opcodes() -> None:
        return None

    def force_compile(func: FuncAny) -> bool:
        return False

    def force_uncompile(func: FuncAny) -> bool:
        return False

    def get_allocator_stats() -> dict[str, int]:
        return {}

    def get_and_clear_inline_cache_stats() -> dict[str, object]:
        return {}

    def get_and_clear_runtime_stats() -> dict[str, object]:
        return {}

    def get_compilation_time() -> int:
        return 0

    def get_compile_after_n_calls() -> int | None:
        return None

    def get_compiled_functions() -> list[FuncAny]:
        return []

    def get_compiled_size(func: FuncAny) -> int:
        return 0

    def get_compiled_spill_stack_size(func: FuncAny) -> int:
        return 0

    def get_compiled_stack_size(func: FuncAny) -> int:
        return 0

    def get_function_compilation_time(func: FuncAny) -> int:
        return 0

    def get_function_hir_opcode_counts(func: FuncAny) -> dict[str, int] | None:
        return {}

    def get_inlined_functions_stats(func: FuncAny) -> dict[str, object]:
        return {}

    def get_jit_list() -> tuple[dict[str, set[str]], dict[str, dict[str, set[int]]]]:
        return ({}, {})

    def get_num_inlined_functions(func: FuncAny) -> int:
        return 0

    def is_enabled() -> bool:
        return False

    def is_hir_inliner_enabled() -> bool:
        return False

    def is_inline_cache_stats_collection_enabled() -> bool:
        return False

    def is_jit_compiled(func: FuncAny) -> bool:
        return False

    def jit_suppress(func: FuncAny) -> FuncAny:
        return func

    def jit_unsuppress(func: FuncAny) -> FuncAny:
        return func

    def lazy_compile(func: FuncAny) -> bool:
        return False

    def mlock_profiler_dependencies() -> None:
        return None

    def multithreaded_compile_test() -> None:
        return None

    def page_in_profiler_dependencies() -> list[str]:
        return []

    def precompile_all(workers: int = 0) -> bool:
        return False

    def read_jit_list(path: str) -> None:
        return None

    def set_max_code_size(max_code_size: int) -> None:
        return None


@contextmanager
def pause(deopt_all: bool = False) -> Generator[None, None, None]:
    """
    Context manager for temporarily pausing the JIT.

    This will disable the JIT from running on new functions, and if you set
    `deopt_all`, will also de-optimize all currently compiled functions to the
    interpreter.  When the JIT is unpaused, the compiled functions will be put
    back.
    """

    prev_enabled = is_enabled()
    if prev_enabled:
        disable(deopt_all=deopt_all)

    try:
        yield
    finally:
        if prev_enabled:
            # Disable the warning from enable() when the JIT is not
            # installed/initialized.
            with catch_warnings():
                simplefilter("ignore")
                enable()



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
from ast import Import, Module
from dataclasses import dataclass
from typing import Any, Sequence

from ..symbols import ModuleScope, Scope, SymbolVisitor


@dataclass
class Flags:
    is_static: bool = False
    is_strict: bool = False

    def merge(self, other: Flags | None) -> Flags:
        if other is None:
            return self

        return Flags(
            is_static=self.is_static or other.is_static,
            is_strict=self.is_strict or other.is_strict,
        )


class BadFlagException(Exception):
    pass


class FlagExtractor(SymbolVisitor):
    is_static: bool
    is_strict: bool

    flag_may_appear: bool
    seen_docstring: bool
    seen_import: bool

    def __init__(self, future_flags: int = 0) -> None:
        super().__init__(future_flags)
        self.is_static = False
        self.is_strict = False

        self.flag_may_appear = True
        self.seen_docstring = False
        self.seen_import = False

    def get_flags(self, node: ast.AST) -> Flags:
        assert isinstance(node, ast.Module)
        self._handle_module(node)
        return Flags(is_static=self.is_static, is_strict=self.is_strict)

    def _handle_module(self, node: Module) -> None:
        for child in node.body:
            match child:
                case ast.Expr(ast.Constant(value)) if (
                    isinstance(value, str) and not self.seen_docstring
                ):
                    self.seen_docstring = True
                case ast.ImportFrom(module) if module == "__future__":
                    pass
                case ast.Constant(_):
                    pass
                case ast.Import(_) as import_node:
                    self._handle_import(import_node)
                case _:
                    self.flag_may_appear = False

    def _handle_import(self, node: Import) -> None:
        for import_ in node.names:
            name = import_.name

            if name not in ("__static__", "__strict__", "__future__"):
                self.flag_may_appear = False
                continue

            if not self.flag_may_appear:
                raise BadFlagException(
                    f"Cinder flag {name} must be at the top of a file"
                )

            if len(node.names) > 1:
                raise BadFlagException(
                    f"{name} flag may not be combined with other imports",
                )

            if import_.asname is not None:
                raise BadFlagException(f"{name} flag may not be aliased")

            match name:
                case "__static__":
                    self.is_static = True
                case "__strict__":
                    self.is_strict = True



# Portions copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

"""Package for compiling Python source code

There are several functions defined at the top level that are imported
from modules contained in the package.

compile(source, filename, mode, flags=None, dont_inherit=None)
    Returns a code object.  A replacement for the builtin compile() function.

compileFile(filename)
    Generates a .pyc file by compiling filename.
"""

import ast
from types import CodeType
from typing import Any

from .pycodegen import CinderCodeGenerator, compile, compile_code, compileFile


def exec_cinder(
    source: str | bytes | ast.Module | ast.Expression | ast.Interactive | CodeType,
    locals: dict[str, Any],
    globals: dict[str, Any],
    modname: str = "<module>",
) -> None:
    if isinstance(source, CodeType):
        code = source
    else:
        code = compile_code(
            source, "<module>", "exec", compiler=CinderCodeGenerator, modname=modname
        )

    exec(code, locals, globals)


__all__ = ("compile", "compile_code", "compileFile", "exec_cinder")



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import linecache
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ast import AST
    from typing import Generator


class TypedSyntaxError(SyntaxError):
    pass


class PerfWarning(Warning):
    def __init__(
        self,
        msg: str,
        filename: str,
        lineno: int,
        offset: int,
        text: str | None,
        end_lineno: int | None = None,
        end_offset: int | None = None,
    ) -> None:
        super().__init__(msg, filename, lineno, offset, text)
        self.msg = msg
        self.filename = filename
        self.lineno = lineno
        self.offset = offset
        self.text = text
        self.end_lineno = end_lineno
        self.end_offset = end_offset


@dataclass
class ErrorLocation:
    filename: str
    node: AST
    lineno: int
    offset: int
    source_line: str | None = None
    end_lineno: int | None = None
    end_offset: int | None = None


def error_location(filename: str, node: AST) -> ErrorLocation:
    # pyre-fixme[16]: `AST` has no attribute `lineno`.
    source_line = linecache.getline(filename, node.lineno)
    # The SyntaxError offset field is 1-indexed:
    # https://docs.python.org/3.10/library/exceptions.html#SyntaxError.offset
    # The AST col_offset field is 0-indexed:
    # https://docs.python.org/3/library/ast.html
    return ErrorLocation(
        filename,
        node,
        node.lineno,
        # pyre-fixme[16]: `AST` has no attribute `col_offset`.
        node.col_offset + 1,
        source_line or None,
        # pyre-fixme[16]: `AST` has no attribute `end_lineno`.
        node.end_lineno + 1 if node.end_lineno is not None else None,
        # pyre-fixme[16]: `AST` has no attribute `end_col_offset`.
        node.end_col_offset + 1 if node.end_col_offset is not None else None,
    )


class ErrorSink:
    throwing = True

    def __init__(self) -> None:
        self.errors: list[TypedSyntaxError] = []
        self.warnings: list[PerfWarning] = []

    def error(self, exception: TypedSyntaxError) -> None:
        raise exception

    def syntax_error(self, msg: str, filename: str, node: AST) -> None:
        errloc = error_location(filename, node)
        self.error(
            TypedSyntaxError(
                msg,
                (
                    errloc.filename,
                    errloc.lineno,
                    errloc.offset,
                    errloc.source_line,
                    errloc.end_lineno,
                    errloc.end_offset,
                ),
            )
        )

    @property
    def has_errors(self) -> bool:
        return len(self.errors) > 0

    @contextmanager
    def error_context(self, filename: str, node: AST) -> Generator[None, None, None]:
        """Add error location context to any TypedSyntaxError raised in with block."""
        try:
            yield
        except TypedSyntaxError as exc:
            if exc.filename is None:
                exc.filename = filename
            if (exc.lineno, exc.offset) == (None, None):
                errloc = error_location(filename, node)
                exc.lineno = errloc.lineno
                exc.offset = errloc.offset
                exc.text = errloc.source_line
                exc.end_lineno = errloc.end_lineno
                exc.end_offset = errloc.end_offset
            self.error(exc)

    def warn(self, warning: PerfWarning) -> None:
        pass

    def perf_warning(self, msg: str, filename: str, node: AST) -> None:
        errloc = error_location(filename, node)
        self.warn(
            PerfWarning(
                msg,
                filename,
                errloc.lineno,
                errloc.offset,
                errloc.source_line,
                errloc.end_lineno,
                errloc.end_offset,
            )
        )

    @property
    def has_warnings(self) -> bool:
        return len(self.warnings) > 0


class CollectingErrorSink(ErrorSink):
    throwing = False

    def error(self, exception: TypedSyntaxError) -> None:
        self.errors.append(exception)

    def warn(self, warning: PerfWarning) -> None:
        self.warnings.append(warning)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from __future__ import annotations

import inspect
from typing import Iterable, Mapping, Sequence, SupportsIndex, Type


def eq_method(self: Enum, other: Enum) -> bool:
    return self.value == other or (
        getattr(other, "value", None) is not None and self.value == other.value
    )


def hash_method(self: Enum) -> int:
    return hash(self.value)


TYPE_NEW: object = type.__new__
SENTINEL = object()


class EnumMeta(type):
    @staticmethod
    def is_hashable(obj: object) -> bool:
        try:
            hash(obj)
            return True
        except TypeError:
            return False

    def __new__(
        metacls,
        classname: str,
        parents: tuple[Type[object], ...],
        dct: Mapping[str, object],
    ) -> type[Enum]:
        attributes = {}
        members = {}

        for name, value in dct.items():
            if (
                name.startswith("_")
                or inspect.isroutine(value)
                or inspect.isdatadescriptor(value)
            ):
                attributes[name] = value
            else:
                members[name] = value

        attributes.update({"__members__": {}, "__reversed_map__": {}})
        if len(members) != len(
            {members[k] for k in members if EnumMeta.is_hashable(members[k])}
        ):
            attributes["__eq__"] = eq_method
            attributes["__hash__"] = hash_method

        klass = super().__new__(metacls, classname, parents, attributes)

        for name, value in members.items():
            option = klass(name=name, value=value)
            # pyre-ignore[16]: __members__ is dynamically defined
            klass.__members__[name] = option
            if EnumMeta.is_hashable(option):
                # pyre-ignore[16]: __reversed_map__ is dynamically defined
                klass.__reversed_map__[option] = option
            if EnumMeta.is_hashable(value):
                klass.__reversed_map__[value] = option
            setattr(klass, name, option)
        return klass

    def __len__(cls) -> int:
        # pyre-ignore[16]: __members__ is dynamically defined
        return len(cls.__members__)

    def __getitem__(cls, attribute: str) -> Enum:
        # pyre-ignore[16]: __members__ is dynamically defined
        return cls.__members__[attribute]

    def __iter__(cls) -> Iterable[Enum]:
        # pyre-ignore[16]: __members__ is dynamically defined
        return iter(cls.__members__.values())

    def __call__(cls, *args: object, **kwargs: object) -> Enum:
        if len(args) == 1:
            attribute = args[0]
            # pyre-ignore[6]: expected str, got object
            return cls._get_by_value(attribute)

        name = kwargs["name"]
        value = kwargs["value"]
        # pyre-ignore[20]: __new__ expects dct
        # pyre-ignore[9]:declared to have type Enum but is used as Type[Enum]
        instance: Enum = cls.__new__(cls, value)
        instance.name = name
        instance.value = value
        return instance

    def _get_by_value(cls, value: str) -> Enum:
        # pyre-ignore[16]: __reversed_map__ is dynamically defined
        res = cls.__reversed_map__.get(value, SENTINEL)
        if res is not SENTINEL:
            return res

        raise ValueError(
            f"Enum type {cls.__name__} has no attribute with value {value!r}"
        )


class Enum(metaclass=EnumMeta):
    def __init__(self, value: object) -> None:
        self.value = value
        self.name: object = None

    def __dir__(self) -> Sequence[str]:
        return ["name", "value"]

    def __str__(self) -> str:
        return f"{type(self).__name__}.{self.name}"

    def __repr__(self) -> str:
        return f"<{type(self).__name__}.{self.name}: {self.value}>"

    def __reduce_ex__(self, proto: SupportsIndex) -> tuple[type[object], tuple[object]]:
        return self.__class__, (self.value,)


class IntEnum(Enum, int):
    pass


class StringEnumMeta(EnumMeta):
    """Like the regular EnumMeta, but parses string/binary inputs to __call__
    as text (to match text literals used in StringEnum)."""

    def _get_by_value(cls, value: str | bytes) -> StringEnum:
        # pyre-ignore[7]: Expected StringEnum, got Enum
        return super()._get_by_value(
            value.decode("utf-8") if isinstance(value, bytes) else value
        )


class StringEnum(Enum, str, metaclass=StringEnumMeta):
    def __str__(self) -> str:
        return f"{self.value}"


def unique(enumeration: type[Enum]) -> type[Enum]:
    """
    Class decorator for enumerations ensuring unique member values
    """
    duplicates = []
    # pyre-ignore[16]: __members__ is dynamically defined
    for name, member in enumeration.__members__.items():
        if name != member.name:
            duplicates.append((name, member.name))
    if duplicates:
        raise ValueError(f"duplicate values found in {enumeration!r}")
    return enumeration



# Portions copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from ast import AST, copy_location, iter_fields
from typing import Sequence, TypeVar

# XXX should probably rename ASTVisitor to ASTWalker
# XXX can it be made even more generic?


TAst = TypeVar("TAst", bound=AST)


class ASTVisitor:
    """Performs a depth-first walk of the AST

    The ASTVisitor is responsible for walking over the tree in the
    correct order.  For each node, it checks the visitor argument for
    a method named 'visitNodeType' where NodeType is the name of the
    node's class, e.g. Class.  If the method exists, it is called
    with the node as its sole argument.

    This is basically the same as the built-in ast.NodeVisitor except
    for the following differences:
        It accepts extra parameters through the visit methods for flowing state
        It uses "visitNodeName" instead of "visit_NodeName"
        It accepts a list to the generic_visit function rather than just nodes
    """

    def __init__(self) -> None:
        self.node: AST | None = None
        self._cache: dict[object, object] = {}

    def generic_visit(self, node: TAst, *args: object) -> object:
        """Called if no explicit visitor function exists for a node."""
        for _field, value in iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, AST):
                        self.visit(item, *args)
            elif isinstance(value, AST):
                self.visit(value, *args)

    def visit(self, node: TAst, *args: object) -> object:
        if not isinstance(node, AST):
            raise TypeError(f"Expected AST node, got {node!r}")

        self.node = node
        klass = node.__class__
        meth = self._cache.get(klass, None)
        if meth is None:
            className = klass.__name__
            meth = getattr(type(self), "visit" + className, type(self).generic_visit)
            self._cache[klass] = meth
        if not args:
            return meth(self, node)
        return meth(self, node, *args)

    def visit_list(self, nodes: Sequence[TAst], *args: object) -> None:
        if not isinstance(nodes, list):
            raise TypeError(f"Expected a list of AST nodes, got {nodes!r}")

        for node in nodes:
            self.visit(node, *args)


class ASTRewriter(ASTVisitor):
    """performs rewrites on the AST, rewriting parent nodes when child nodes
    are replaced."""

    @staticmethod
    def update_node(node: TAst, **replacement: object) -> TAst:
        res = node
        for name, val in replacement.items():
            existing = getattr(res, name)
            if existing is val:
                continue

            if node is res:
                res = ASTRewriter.clone_node(node)

            setattr(res, name, val)
        return res

    @staticmethod
    def clone_node(node: TAst) -> TAst:
        attrs = []
        for name in node._fields:
            attr = getattr(node, name, None)
            if isinstance(attr, list):
                attr = list(attr)
            attrs.append(attr)

        new = type(node)(*attrs)
        return copy_location(new, node)

    def skip_field(self, node: TAst, field: str) -> bool:
        return False

    def generic_visit(self, node: TAst, *args: object) -> TAst:
        ret_node = node
        for field, old_value in iter_fields(node):
            if self.skip_field(node, field):
                continue

            if isinstance(old_value, AST):
                new_value = self.visit(old_value, *args)
            elif isinstance(old_value, list):
                new_value = self.walk_list(old_value, *args)
            else:
                continue

            assert new_value is not None, (
                f"can't remove AST nodes that aren't part of a list {old_value!r}"
            )
            if new_value is not old_value:
                if ret_node is node:
                    ret_node = self.clone_node(node)

                setattr(ret_node, field, new_value)

        return ret_node

    def walk_list(self, values: Sequence[object], *args: object) -> Sequence[object]:
        """
        Like visit_list(), but it also walks values returned by ast.iter_fields()
        so it has to handle non-AST node values.
        """

        new_values = []
        changed = False

        for value in values:
            if not isinstance(value, AST):
                new_values.append(value)
                continue

            new_value = self.visit(value, *args)
            if new_value is not None:
                new_values.append(new_value)
                changed = True

        # Reuse the existing list when possible.
        return new_values if changed else values


class ExampleASTVisitor(ASTVisitor):
    """Prints examples of the nodes that aren't visited

    This visitor-driver is only useful for development, when it's
    helpful to develop a visitor incrementally, and get feedback on what
    you still have to do.
    """

    VERBOSE: int = 0

    examples: dict[object, object] = {}

    def visit(self, node: TAst, *args: object) -> object:
        self.node = node
        meth = self._cache.get(node.__class__, None)
        className = node.__class__.__name__
        if meth is None:
            meth = getattr(self, "visit" + className, 0)
            self._cache[node.__class__] = meth
        if self.VERBOSE > 1:
            print("visit", className, meth and meth.__name__ or "")
        if meth:
            meth(node, *args)
        elif self.VERBOSE > 0:
            klass = node.__class__
            if klass not in self.examples:
                self.examples[klass] = klass
                print()
                print(self)
                print(klass)
                for attr in dir(node):
                    if attr[0] != "_":
                        print("\t", "%-12.12s" % attr, getattr(node, attr))
                print()
            return self.default(node, *args)

    def default(self, node: TAst, *args: object) -> TAst:
        return node


# XXX this is an API change


def dumpNode(node: object) -> None:
    print(node.__class__)
    for attr in dir(node):
        if attr[0] != "_":
            print("\t", "%-10.10s" % attr, getattr(node, attr))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict
"""High-performance Python runtime extensions."""

from __future__ import annotations

import gc
import platform
import sys
from os import environ

# ============================================================================
# Note!
#
# At Meta, this module is currently loaded as part of Lib/site.py in an attempt
# to get benefits from using it as soon as possible.  However
# Lib/test/test_site.py will assert that site.py does not import too many
# modules.  Be careful with adding import statements here.
#
# The plan is to move applications over to using an explicit initialization
# step rather than Lib/site.py.  Once that is done we can add all the imports
# we want here.
# ============================================================================


_import_error: ImportError | None = None


def is_supported_runtime() -> bool:
    """
    Check that the current Python runtime will be able to load the _cinderx
    native extension.
    """

    if sys.platform not in ("darwin", "linux"):
        return False

    version = (sys.version_info.major, sys.version_info.minor)
    if version == (3, 14) or version == (3, 15):
        # Can't load the native extension if the GIL is forcibly being disabled.  The
        # native extension doesn't support free-threading properly yet.
        return environ.get("PYTHON_GIL") != "0"
    if version == (3, 12):
        return "+meta" in sys.version
    if version == (3, 10):
        return "+cinder" in sys.version
    return False


try:
    # Currently if we try to import _cinderx on runtimes without our internal patches
    # the import will crash.  This is meant to go away in the future.
    if not is_supported_runtime():
        raise ImportError(
            f"The _cinderx native extension is not supported for Python version '{sys.version}' on platform '{sys.platform}'"
        )

    # pyre-ignore[21]: _cinderx is not a real cpp_python_extension() yet.
    from _cinderx import (  # noqa: F401
        _compile_perf_trampoline_pre_fork,
        _is_compile_perf_trampoline_pre_fork_enabled,
        async_cached_classproperty,
        async_cached_property,
        cached_classproperty,
        cached_property,
        cached_property_with_descr,
        clear_caches,
        clear_classloader_caches,
        disable_parallel_gc,
        enable_parallel_gc,
        freeze_type,
        get_parallel_gc_settings,
        has_parallel_gc,
        immortalize_heap,
        is_immortal,
        strict_module_patch,
        strict_module_patch_delete,
        strict_module_patch_enabled,
        StrictModule,
        watch_sys_modules,
    )

    if sys.version_info < (3, 11):
        # In 3.12+ use the versions in the polyfill cinder library instead.
        from _cinderx import (
            _get_entire_call_stack_as_qualnames_with_lineno,
            _get_entire_call_stack_as_qualnames_with_lineno_and_frame,
            clear_all_shadow_caches,
        )

    if sys.version_info >= (3, 12):
        from _cinderx import delay_adaptive, get_adaptive_delay, set_adaptive_delay

except ImportError as e:
    if "undefined symbol:" in str(e):
        # If we're on a dev build report this as an error, otherwise muddle along with alternative definitions
        # on unsupported Python's.
        from os.path import dirname, exists, join

        if exists(join(dirname(__file__), ".dev_build")):
            raise ImportError(
                "The _cinderx native extension is not available due to a missing symbol. This is likely a bug you introduced.  "
                "Please ensure that the cinderx kernel is being used."
            ) from e
    _import_error = e

    def _compile_perf_trampoline_pre_fork() -> None:
        pass

    def _get_entire_call_stack_as_qualnames_with_lineno() -> list[tuple[str, int]]:
        return []

    def _get_entire_call_stack_as_qualnames_with_lineno_and_frame() -> list[
        tuple[str, int, object]
    ]:
        return []

    def _is_compile_perf_trampoline_pre_fork_enabled() -> bool:
        return False

    from asyncio import AbstractEventLoop, Future
    from typing import (
        Awaitable,
        Callable,
        Dict,
        final,
        Generator,
        Generic,
        List,
        NoReturn,
        Optional,
        overload,
        Tuple,
        Type,
        TYPE_CHECKING,
        TypeVar,
    )

    _TClass = TypeVar("_TClass")
    _TReturnType = TypeVar("_TReturnType")

    @final
    class NoValueSet:
        pass

    NO_VALUE_SET = NoValueSet()

    class _BaseCachedProperty(Generic[_TClass, _TReturnType]):
        fget: Callable[[_TClass], _TReturnType]
        __name__: str

        def __init__(
            self,
            f: Callable[[_TClass], _TReturnType],
            slot: Optional[Descriptor[_TReturnType]] = None,
        ) -> None:
            self.fget: Callable[[_TClass], _TReturnType] = f
            self.__name__ = f.__name__
            self.__doc__: str | None = f.__doc__
            self.slot = slot

        @overload
        def __get__(
            self, obj: None, cls: Type[_TClass]
        ) -> _BaseCachedProperty[_TClass, _TReturnType]: ...

        @overload
        def __get__(self, obj: _TClass, cls: Type[_TClass]) -> _TReturnType: ...

        def __get__(
            self, obj: Optional[_TClass], cls: Type[_TClass]
        ) -> _BaseCachedProperty[_TClass, _TReturnType] | _TReturnType:
            if obj is None:
                return self
            slot = self.slot
            if slot is not None:
                try:
                    res = slot.__get__(obj, cls)
                except AttributeError:
                    res = self.fget(obj)
                    slot.__set__(obj, res)
                return res

            result = self.fget(obj)
            obj.__dict__[self.__name__] = result
            return result

    class _AsyncLazyValueState:
        NotStarted = 0
        Running = 1
        Done = 2

    _T = TypeVar("_T", covariant=True)
    _TParams = TypeVar("_TParams")

    # noqa: F401
    import asyncio

    class _AsyncLazyValue(Awaitable[_T]):
        """
        This is a low-level class used mainly for two things:
        * It helps to avoid calling a coroutine multiple times, by caching the
        result of a previous call
        * It ensures that the coroutine is called only once

        _AsyncLazyValue has well defined cancellation behavior in these cases:

        1. When we have a single task stack (call stack for you JS folks), which is
        awaiting on the AsyncLazyValue
        -> In this case, we mimic the behavior of a normal await. i.e: If the
            task stack gets cancelled, we cancel the coroutine (by raising a
            CancelledError in the underlying future)

        2. When we have multiple task stacks awaiting on the future.
        We have two sub cases here.

        2.1. The initial task stack (which resulted in an await of the coroutine)
                gets cancelled.
                -> In this case, we cancel the coroutine, and all the tasks depending
                on it. If we don't do that, we'd have to implement retry logic,
                which is a bad idea in such low level code. Even if we do implement
                retries, there's no guarantee that they would succeed, so it's better
                to just fail here.

                Also, the number of times this happens is very small (I don't have
                data to prove it, but qualitative arguments suggest this is the
                case).

        2.2. One of the many task stacks gets cancelled (but not the one which ended
            up awaiting the coroutine)
            -> In this case, we just allow the task stack to be cancelled, but
                the rest of them are processed without being affected.
        """

        def __init__(
            self,
            # pyre-fixme[31]: Expression `typing.Callable[(_TParams,
            #  typing.Awaitable[_T])]` is not a valid type.
            coro_func: Callable[_TParams, Awaitable[_T]],
            # pyre-fixme[11]: Annotation `args` is not defined as a type.
            *args: _TParams.args,
            # pyre-fixme[11]: Annotation `kwargs` is not defined as a type.
            **kwargs: _TParams.kwargs,
        ) -> None:
            global asyncio
            # pyre-fixme[31]: Expression `typing.Optional[typing.Callable[(_TParams,
            #  typing.Awaitable[_T])]]` is not a valid type.
            self.coro_func: Optional[Callable[_TParams, Awaitable[_T]]] = coro_func
            self.args: Tuple[object, ...] = args
            self.kwargs: Dict[str, object] = kwargs
            self.state: int = _AsyncLazyValueState.NotStarted
            self.res: Optional[_T] = None
            self._futures: List[Future] = []
            self._awaiting_tasks = 0

        async def _async_compute(self) -> _T:
            futures = self._futures
            try:
                coro_func = self.coro_func
                # lint-fixme: NoAssertsRule
                assert coro_func is not None
                self.res = res = await coro_func(*self.args, **self.kwargs)

                self.state = _AsyncLazyValueState.Done

                # pyre-fixme[1001]: Awaitable assigned to `value` is never awaited.
                for value in futures:
                    if not value.done():
                        value.set_result(self.res)

                self.args = ()
                self.kwargs.clear()
                del self._futures[:]
                self.coro_func = None

                return res

            except (Exception, asyncio.CancelledError) as e:
                # pyre-fixme[1001]: Awaitable assigned to `value` is never awaited.
                for value in futures:
                    if not value.done():
                        value.set_exception(e)
                self._futures = []
                self.state = _AsyncLazyValueState.NotStarted
                raise

        def _get_future(self, loop: Optional[AbstractEventLoop]) -> Future:
            if loop is None:
                loop = asyncio.get_event_loop()
            f = asyncio.Future(loop=loop)
            self._futures.append(f)
            self._awaiting_tasks += 1
            return f

        def __iter__(self) -> _AsyncLazyValue[_T]:
            return self

        def __next__(self) -> NoReturn:
            raise StopIteration(self.res)

        def __await__(self) -> Generator[None, None, _T]:
            if self.state == _AsyncLazyValueState.Done:
                # pyre-ignore[7]: Expected `Generator[None, None, Variable[_T](covariant)]`
                # but got `_AsyncLazyValue[Variable[_T](covariant)]`.
                return self
            elif self.state == _AsyncLazyValueState.Running:
                c = self._get_future(None)
                return c.__await__()
            else:
                self.state = _AsyncLazyValueState.Running
                c = self._async_compute()
                return c.__await__()

        def as_future(self, loop: AbstractEventLoop) -> Future:
            if self.state == _AsyncLazyValueState.Done:
                f = asyncio.Future(loop=loop)
                f.set_result(self.res)
                return f
            elif self.state == _AsyncLazyValueState.Running:
                return self._get_future(loop)
            else:
                if loop is None:
                    loop = asyncio.get_event_loop()
                t = loop.create_task(self._async_compute())
                self.state = _AsyncLazyValueState.Running
                # pyre-ignore[16]: Undefined attribute `asyncio.tasks.Task`
                # has no attribute `_source_traceback`.
                if t._source_traceback:
                    del t._source_traceback[-1]
                # pyre-fixme[7]: Expected `Future[Any]` but got `Task[_T]`.
                return t

    _TAwaitableReturnType = TypeVar("_TAwaitableReturnType")

    class async_cached_property(
        Generic[_TAwaitableReturnType, _TClass],
        _BaseCachedProperty[_TClass, Awaitable[_TAwaitableReturnType]],
    ):
        def __init__(
            self,
            f: Callable[[_TClass], _TReturnType],
            slot: Optional[Descriptor[_TReturnType]] = None,
        ) -> None:
            super().__init__(f, slot)

        def __get__(
            self, obj: Optional[_TClass], cls: Type[_TClass]
        ) -> (
            _BaseCachedProperty[_TClass, Awaitable[_TAwaitableReturnType]]
            | Awaitable[_TAwaitableReturnType]
        ):
            if obj is None:
                return self

            slot = self.slot
            if slot is not None:
                try:
                    res = slot.__get__(obj, cls)
                except AttributeError:
                    res = _AsyncLazyValue(self.fget, obj)
                    slot.__set__(obj, res)
                return res

            lazy_value = _AsyncLazyValue(self.fget, obj)
            setattr(obj, self.__name__, lazy_value)
            return lazy_value

    class async_cached_classproperty(
        Generic[_TAwaitableReturnType, _TClass],
        _BaseCachedProperty[Type[_TClass], Awaitable[_TAwaitableReturnType]],
    ):
        def __init__(
            self,
            f: Callable[[_TClass], Awaitable[_TAwaitableReturnType]],
            slot: Optional[Descriptor[Awaitable[_TAwaitableReturnType]]] = None,
        ) -> None:
            super().__init__(f, slot)
            self._value: NoValueSet | Awaitable[_TAwaitableReturnType] = NO_VALUE_SET

        def __get__(
            self, obj: Optional[_TClass], cls: Type[_TClass]
        ) -> Awaitable[_TAwaitableReturnType]:
            lazy_value = self._value
            if not isinstance(lazy_value, NoValueSet):
                return lazy_value
            self._value = lazy_value = _AsyncLazyValue(self.fget, cls)
            return lazy_value

    class cached_classproperty(_BaseCachedProperty[Type[_TClass], _TReturnType]):
        def __init__(
            self,
            f: Callable[[_TClass], _TReturnType],
            slot: Optional[Descriptor[_TReturnType]] = None,
        ) -> None:
            super().__init__(f, slot)
            self._value: NoValueSet | _TReturnType = NO_VALUE_SET

        def __get__(self, obj: Optional[_TClass], cls: Type[_TClass]) -> _TReturnType:
            result = self._value
            if not isinstance(result, NoValueSet):
                return result
            self._value = result = self.fget(cls)
            return result

    _TClass = TypeVar("_TClass")
    _TReturnType = TypeVar("_TReturnType")

    if TYPE_CHECKING:
        from abc import ABC

        @final
        class Descriptor(ABC, Generic[_TReturnType]):
            __name__: str
            __objclass__: Type[object]

            def __get__(
                self, inst: object, ctx: Optional[Type[object]] = None
            ) -> _TReturnType: ...

            def __set__(self, inst: object, value: _TReturnType) -> None:
                pass

            def __delete__(self, inst: object) -> None:
                pass

    class cached_property(_BaseCachedProperty[_TClass, _TReturnType]):
        def __init__(
            self,
            f: Callable[[_TClass], _TReturnType],
            slot: Optional[Descriptor[_TReturnType]] = None,
        ) -> None:
            super().__init__(f, slot)
            if slot is not None:
                if (
                    type(self) is not cached_property
                    and type(self) is not cached_property_with_descr
                ):
                    raise TypeError(
                        "slot can't be used with subtypes of cached_property"
                    )
                # pyre-ignore[4]: Missing attribute annotation for __class__
                self.__class__ = cached_property_with_descr

    class cached_property_with_descr(cached_property[_TClass, _TReturnType]):
        def __set__(self, inst: object, value: _TReturnType) -> None:
            slot = self.slot
            if slot is not None:
                slot.__set__(inst, value)
            else:
                setattr(inst, self.__name__, value)

        def __delete__(self, inst: object) -> None:
            slot = self.slot
            if slot is not None:
                slot.__delete__(inst)
            else:
                delattr(inst, self.__name__)

    if sys.version_info < (3, 11):

        def clear_all_shadow_caches() -> None:
            pass

    def clear_caches() -> None:
        pass

    def clear_classloader_caches() -> None:
        pass

    def disable_parallel_gc() -> None:
        pass

    def enable_parallel_gc(min_generation: int = 2, num_threads: int = 0) -> None:
        raise RuntimeError(
            "No Parallel GC support because _cinderx did not load correctly"
        )

    def freeze_type(ty: object) -> object:
        return ty

    def get_parallel_gc_settings() -> dict[str, int] | None:
        return None

    def has_parallel_gc() -> bool:
        return False

    def immortalize_heap() -> None:
        pass

    def is_immortal(obj: object) -> bool:
        raise RuntimeError(
            "Can't answer whether an object is mortal or immortal from Python code"
        )

    def strict_module_patch(mod: object, name: str, value: object) -> None:
        pass

    def strict_module_patch_delete(mod: object, name: str) -> None:
        pass

    def strict_module_patch_enabled(mod: object) -> bool:
        return False

    class StrictModule:
        def __init__(self, d: dict[str, object], b: bool) -> None:
            pass

    def watch_sys_modules() -> None:
        pass


def maybe_enable_parallel_gc() -> None:
    """Conditionally enable parallel GC based on environment variables."""
    is_parallel_gc_enabled = environ.get("PARALLEL_GC_ENABLED", "0") == "1"
    if not has_parallel_gc() or not is_parallel_gc_enabled:
        return
    thresholds = gc.get_threshold()
    parallel_gc_threshold_gen0 = int(
        environ.get("PARALLEL_GC_THRESHOLD_GEN0", thresholds[0])
    )
    parallel_gc_threshold_gen1 = int(
        environ.get("PARALLEL_GC_THRESHOLD_GEN1", thresholds[1])
    )
    parallel_gc_threshold_gen2 = int(
        environ.get("PARALLEL_GC_THRESHOLD_GEN2", thresholds[2])
    )
    gc.set_threshold(
        parallel_gc_threshold_gen0,
        parallel_gc_threshold_gen1,
        parallel_gc_threshold_gen2,
    )

    parallel_gc_num_threads = int(environ.get("PARALLEL_GC_NUM_THREADS", "0"))
    parallel_gc_min_generation = int(environ.get("PARALLEL_GC_MIN_GENERATION", "2"))

    enable_parallel_gc(
        min_generation=parallel_gc_min_generation,
        num_threads=parallel_gc_num_threads,
    )


_is_init: bool = False


def init() -> None:
    """Initialize CinderX."""
    global _is_init

    # Failed to import _cinderx, nothing to initialize.
    if _import_error is not None:
        return

    # Already initialized.
    if _is_init:
        return

    maybe_enable_parallel_gc()

    _is_init = True


def is_initialized() -> bool:
    """
    Check if the cinderx extension has been properly initialized.
    """
    return _is_init


def get_import_error() -> ImportError | None:
    """
    Get the ImportError that occurred when _cinderx was imported, if there was
    an error.
    """
    return _import_error


init()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

import abc
import ctypes
import dis
import functools
import importlib
import multiprocessing
import os.path
import sys
import tempfile
import types
import unittest
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Coroutine, Generator, Sequence, TypeVar

import cinderx
import cinderx.jit

try:
    import cinder

    def hasCinderX() -> bool:
        return True

except ImportError:

    def hasCinderX() -> bool:
        return False


# String encoding to use for subprocesses.
ENCODING: str = sys.stdout.encoding or sys.getdefaultencoding()

# Hack to allow subprocesses to find where the cinderx module is.
CINDERX_PATH: str = os.path.dirname(os.path.dirname(cinderx.__file__))


def subprocess_env() -> dict[str, str]:
    # We have CINDERX_PATH here over the rest of sys.path because our cinderx/opcode.py
    # library is generated on a per-version basis and ends up being thrown into cinderx
    # at build time. When we run tests with sys.path we can end up with PythonLib on the
    # path first and we can pick up the version which doesn't have it injected.
    return {
        "PYTHONPATH": CINDERX_PATH + os.path.pathsep + os.path.pathsep.join(sys.path)
    }


def get_cinderjit_xargs() -> list[str]:
    args = []
    for k, v in sys._xoptions.items():
        if not k.startswith("jit"):
            continue
        elif v is True:
            args.extend(["-X", k])
        else:
            args.extend(["-X", f"{k}={v}"])
    return args


TYield = TypeVar("TYield")
TSend = TypeVar("TSend")
TReturn = TypeVar("TReturn")


def get_await_stack(
    coro: Coroutine[TYield, TSend, TReturn],
) -> list[Coroutine[TYield, TSend, TReturn]]:
    """Return the chain of coroutines reachable from coro via its awaiter"""

    stack = []
    awaiter = cinder._get_coro_awaiter(coro)
    while awaiter is not None:
        stack.append(awaiter)

        # pyre-ignore[1001]
        awaiter = cinder._get_coro_awaiter(awaiter)
    return stack


def verify_stack(
    testcase: unittest.TestCase, stack: Sequence[str], expected: Sequence[str]
) -> None:
    n = len(expected)
    frames = stack[-n:]
    testcase.assertEqual(len(frames), n, "Callstack had less frames than expected")

    for actual, exp in zip(frames, expected):
        testcase.assertTrue(
            actual.endswith(exp),
            f"The actual frame {actual} doesn't refer to the expected function {exp}",
        )


def compiles_after_one_call() -> bool:
    """
    Check if CinderX will automatically compile functions after they are called once.
    """
    return cinderx.jit.get_compile_after_n_calls() == 0


_FT = TypeVar("_FT", bound=Callable[..., object])


# pyre-ignore[34]: Type variable isn't present in parameters
def passAlways(reason: str) -> Callable[[_FT], _FT]:
    """
    Force a test to always pass.
    Useful when `skip` is not desired
    (e.g. intentionally skipping tests that shouldn't be deleted)
    """

    def decorator(test_item: object) -> object:
        if isinstance(test_item, type):
            # apply this decorator to all "test_" methods of the test case
            for attr_name in dir(test_item):
                if attr_name.startswith("test_"):
                    attr = getattr(test_item, attr_name)
                    setattr(test_item, attr_name, passAlways(attr))

        else:

            @functools.wraps(test_item)
            def pass_wrapper(*args: object, **kwargs: object) -> None:
                return

            test_item = pass_wrapper

        test_item.__unittest_skip_why__ = reason
        return test_item

    if isinstance(reason, types.FunctionType):
        test_item = reason
        reason = ""
        return decorator(test_item)
    # pyre-ignore[7]: bad return type
    return decorator


# pyre-ignore[34]: Type variable isn't present in parameters
def passIf(condition: object, reason: str) -> Callable[[_FT], _FT]:
    """
    Force a test to pass if the condition is true.
    """
    if condition:
        return passAlways(reason)
    return lambda obj: obj


# pyre-ignore[34]: Type variable isn't present in parameters
def passUnless(condition: object, reason: str) -> Callable[[_FT], _FT]:
    """
    Force a test to pass unless the condition is true.
    """
    if not condition:
        return passAlways(reason)
    return lambda obj: obj


def skip_if_jit(reason: str) -> Callable[[Callable[..., None]], Callable[..., None]]:
    return passIf(cinderx.jit.is_enabled(), reason)


def skip_unless_jit(
    reason: str,
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    return passUnless(cinderx.jit.is_enabled(), reason)


def skip_unless_lazy_imports(
    reason: str = "Depends on Lazy Imports being enabled",
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    return passUnless(hasattr(importlib, "set_lazy_imports"), reason)


TRet = TypeVar("TRet")


def failUnlessJITCompiled(func: Callable[..., TRet]) -> Callable[..., TRet]:
    """
    Fail a test if the JIT is enabled but the test body wasn't JIT-compiled.
    """
    if not cinderx.jit.is_enabled():
        return func

    try:
        # force_compile raises a RuntimeError if compilation fails. If it does,
        # defer raising an exception to when the decorated function runs.
        cinderx.jit.force_compile(func)
    except RuntimeError as re:
        if re.args == ("PYJIT_RESULT_NOT_ON_JITLIST",):
            # We generally only run tests with a jitlist under
            # Tools/scripts/jitlist_bisect.py. In that case, we want to allow
            # the decorated function to run under the interpreter to determine
            # if it's the function the JIT is handling incorrectly.
            return func

        # re is cleared at the end of the except block but we need the value
        # when wrapper() is eventually called.
        exc: RuntimeError = re

        def wrapper(*args: ...) -> None:
            raise RuntimeError(
                f"JIT compilation of {func.__qualname__} failed with {exc}"
            )

        return wrapper

    return func


def fail_if_deopt(func: Callable[..., TRet]) -> Callable[..., TRet]:
    """
    Raise a RuntimeException if _any_ deopts occur during execution of the
    wrapped function. Note deopts occuring in nested function calls will also
    trigger this. Also, execution will run to completion - it won't stop at the
    point a deopt occurs.
    """

    if not cinderx.jit.is_enabled():
        return func

    def wrapper(*args: ..., **kwargs: ...) -> TRet:
        cinderx.jit.get_and_clear_runtime_stats()
        r = func(*args, **kwargs)
        # pyre-ignore[6]
        if len(deopts := cinderx.jit.get_and_clear_runtime_stats()["deopt"]):
            raise RuntimeError(f"Deopt occured {deopts}")
        return r

    wrapper.inner_function = func

    return wrapper


def is_asan_build() -> bool:
    try:
        ctypes.pythonapi.__asan_init
        return True
    except AttributeError:
        return False


# This is long because ASAN + JIT + subprocess + the Python compiler can be
# pretty slow in CI.
SUBPROCESS_TIMEOUT_SEC = 100 if is_asan_build() else 5


@contextmanager
def temp_sys_path() -> Generator[Path, None, None]:
    with tempfile.TemporaryDirectory() as tmpdir:
        _orig_sys_modules = sys.modules
        sys.modules = _orig_sys_modules.copy()
        _orig_sys_path = sys.path[:]
        sys.path.insert(0, tmpdir)
        try:
            yield Path(tmpdir)
        finally:
            sys.path[:] = _orig_sys_path
            sys.modules = _orig_sys_modules


class _ExceptionResult:
    def __init__(self, exc: Exception) -> None:
        self.exc = exc


def run_in_subprocess(func: Callable[..., TRet]) -> Callable[..., TRet]:
    """
    Run a function in a subprocess.  This enables modifying process state in a
    test without affecting other test functions.
    """

    queue: multiprocessing.Queue = multiprocessing.Queue()

    def wrapper(queue: multiprocessing.Queue, *args: object) -> None:
        try:
            result = func(*args)
            queue.put(result, timeout=SUBPROCESS_TIMEOUT_SEC)
        except Exception as e:
            queue.put(_ExceptionResult(e), timeout=SUBPROCESS_TIMEOUT_SEC)

    def wrapped(*args: object) -> TRet:
        fork = multiprocessing.get_context("fork")
        p = fork.Process(target=wrapper, args=(queue, *args))
        p.start()
        value = queue.get(timeout=SUBPROCESS_TIMEOUT_SEC)
        p.join(timeout=SUBPROCESS_TIMEOUT_SEC)
        if isinstance(value, _ExceptionResult):
            raise value.exc
        return value

    return wrapped


class AssertBytecodeContainsMixin(abc.ABC):
    @abc.abstractmethod
    def assertIn(
        self, expected: object, actual: Sequence[object], msg: str | None = None
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def assertTrue(self, expr: object, msg: str | None = None) -> None:
        raise NotImplementedError

    def assertBytecodeContains(
        self,
        func: object,
        expected_opcode: str,
        expected_oparg: int | None = None,
    ) -> None:
        try:
            # pyre-ignore[16] - for things wrapped by fail_if_deopt()
            inner_function = func.inner_function
        except AttributeError:
            pass
        else:
            func = inner_function

        bytecode_instructions = dis.get_instructions(func)

        if expected_oparg is None:
            opcodes = [instr.opname for instr in bytecode_instructions]
            self.assertIn(
                expected_opcode,
                opcodes,
                f"{expected_opcode} opcode should be present in {func.__name__} bytecode",
            )
        else:
            matching_instructions = [
                instr
                for instr in bytecode_instructions
                if instr.opname == expected_opcode and instr.arg == expected_oparg
            ]
            self.assertTrue(
                len(matching_instructions) > 0,
                f"{expected_opcode} opcode with oparg {expected_oparg} should be present in {func.__name__} bytecode",
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

try:  # ensure all imports in this module are eager, to avoid cycles
    import _imp
    import builtins
    import importlib
    import marshal
    import os
    import sys
    from enum import Enum

    # pyre-ignore[21]: typeshed doesn't know about this
    from importlib import _bootstrap, _pack_uint32

    # pyre-ignore[21]: typeshed doesn't know about this
    from importlib._bootstrap_external import (
        _classify_pyc,
        _compile_bytecode,
        _validate_hash_pyc,
        _validate_timestamp_pyc,
    )
    from importlib.abc import Loader
    from importlib.machinery import (
        BYTECODE_SUFFIXES,
        EXTENSION_SUFFIXES,
        ExtensionFileLoader,
        FileFinder,
        ModuleSpec,
        SOURCE_SUFFIXES,
        SourceFileLoader,
        SourcelessFileLoader,
    )
    from importlib.util import cache_from_source, MAGIC_NUMBER
    from io import BytesIO
    from os import getenv, makedirs
    from os.path import dirname, isdir
    from py_compile import (
        _get_default_invalidation_mode,
        PycInvalidationMode,
        PyCompileError,
    )
    from types import CodeType, ModuleType
    from typing import Callable, cast, Collection, final, Iterable, Mapping

    from _cinderx import StrictModule, watch_sys_modules
    from cinderx.static import install_sp_audit_hook

    from ..consts import CI_CO_STATICALLY_COMPILED
    from .common import (
        DEFAULT_STUB_PATH,
        FIXED_MODULES,
        MAGIC_NUMBER as STRICT_MAGIC_NUMBER,
    )
    from .compiler import Compiler, Dependencies, SourceInfo, TIMING_LOGGER_TYPE
    from .flag_extractor import Flags
except BaseException:
    raise


_MAGIC_STRICT_OR_STATIC: bytes = (STRICT_MAGIC_NUMBER + 2**15).to_bytes(
    2, "little"
) + b"\r\n"
# We don't actually need to increment anything here, because the strict modules
# AST rewrite has no impact on pycs for non-strict modules. So we just always
# use two zero bytes. This simplifies generating "fake" strict pycs for
# known-not-to-be-strict third-party modules.
_MAGIC_NEITHER_STRICT_NOR_STATIC: bytes = (0).to_bytes(2, "little") + b"\r\n"
_MAGIC_LEN: int = len(_MAGIC_STRICT_OR_STATIC)


@final
class _PatchState(Enum):
    """Singleton used for tracking values which have not yet been patched."""

    Patched = 1
    Deleted = 2
    Unpatched = 3


# Unfortunately module passed in could be a mock object,
# which also has a `patch` method that clashes with the StrictModule method.
# Directly get the function to avoid name clash.


def _set_patch(module: StrictModule, name: str, value: object) -> None:
    type(module).patch(module, name, value)


def _del_patch(module: StrictModule, name: str) -> None:
    type(module).patch_delete(module, name)


@final
class StrictModuleTestingPatchProxy:
    """Provides a proxy object which enables patching of a strict module if the
    module has been loaded with the StrictSourceWithPatchingFileLoader.  The
    proxy can be used as a context manager in which case exiting the with block
    will result in the patches being disabled.  The process will be terminated
    if the patches are not unapplied and the proxy is deallocated."""

    def __init__(self, module: StrictModule) -> None:
        object.__setattr__(self, "module", module)
        object.__setattr__(self, "_patches", {})
        object.__setattr__(self, "__name__", module.__name__)
        object.__setattr__(
            self, "_final_constants", getattr(module, "__final_constants__", ())
        )
        # pyre-ignore[16]: pyre doesn't understand properties well enough
        if not type(module).__patch_enabled__.__get__(module, type(module)):
            raise ValueError(f"strict module {module} does not allow patching")

    def __setattr__(self, name: str, value: object) -> None:
        patches = object.__getattribute__(self, "_patches")
        prev_patched = patches.get(name, _PatchState.Unpatched)
        module = object.__getattribute__(self, "module")
        final_constants = object.__getattribute__(self, "_final_constants")
        if name in final_constants:
            raise AttributeError(
                f"Cannot patch Final attribute `{name}` of module `{module.__name__}`"
            )
        if value is prev_patched:
            # We're restoring the previous value
            del patches[name]
        elif prev_patched is _PatchState.Unpatched:
            # We're overwriting a value
            # only set patches[name] when name is patched for the first time
            patches[name] = getattr(module, name, _PatchState.Patched)

        if value is _PatchState.Deleted:
            _del_patch(module, name)
        else:
            _set_patch(module, name, value)

    def __delattr__(self, name: str) -> None:
        StrictModuleTestingPatchProxy.__setattr__(self, name, _PatchState.Deleted)

    def __getattribute__(self, name: str) -> object:
        res = getattr(object.__getattribute__(self, "module"), name)
        return res

    def __enter__(self) -> StrictModuleTestingPatchProxy:
        return self

    def __exit__(self, *excinfo: object) -> None:
        StrictModuleTestingPatchProxy.cleanup(self)

    def cleanup(self, ignore: Collection[str] | None = None) -> None:
        patches = object.__getattribute__(self, "_patches")
        module = object.__getattribute__(self, "module")
        for name, value in list(patches.items()):
            if ignore and name in ignore:
                del patches[name]
                continue
            if value is _PatchState.Patched:
                # value is patched means that module originally
                # does not contain this field.
                try:
                    _del_patch(module, name)
                except AttributeError:
                    pass
                finally:
                    del patches[name]
            else:
                setattr(self, name, value)
        assert not patches

    def __del__(self) -> None:
        patches = object.__getattribute__(self, "_patches")
        if patches:
            print(
                "Patch(es)",
                ", ".join(patches.keys()),
                "failed to be detached from strict module",
                "'" + object.__getattribute__(self, "module").__name__ + "'",
                file=sys.stderr,
            )
            # There's a test that depends on this being mocked out.
            os.abort()


__builtins__: ModuleType


class StrictBytecodeError(ImportError):
    pass


def classify_strict_pyc(
    data: bytes, name: str, exc_details: dict[str, str]
) -> tuple[int, bool]:
    # pyre-ignore[16]: typeshed doesn't know about this
    flags = _classify_pyc(data[_MAGIC_LEN:], name, exc_details)
    magic = data[:_MAGIC_LEN]
    if magic == _MAGIC_NEITHER_STRICT_NOR_STATIC:
        strict_or_static = False
    elif magic == _MAGIC_STRICT_OR_STATIC:
        strict_or_static = True
    else:
        raise StrictBytecodeError(
            f"Bad magic number {magic!r} in {exc_details['path']}"
        )
    return (flags, strict_or_static)


# A dependency is (name, mtime_and_size_or_hash)
DependencyTuple = tuple[str, bytes]


def validate_dependencies(
    compiler: Compiler, deps: tuple[DependencyTuple, ...], hash_based: bool
) -> None:
    """Raise ImportError if any dependency has changed."""
    for modname, data in deps:
        # empty invalidation data means non-static
        was_static = bool(data)
        source_info = compiler.get_source(
            modname, need_contents=(hash_based or not was_static)
        )
        if source_info is None:
            if was_static:
                raise ImportError(f"{modname} is missing")
            continue
        if was_static:
            expected_data = get_dependency_data(source_info, hash_based)
            if data != expected_data:
                raise ImportError(f"{modname} has changed")
        else:
            assert source_info.source is not None
            # if a previously non-static dependency is now found and has the text
            # "import __static__" in it, we invalidate
            if b"import __static__" in source_info.source:
                raise ImportError(f"{modname} may now be static")


def get_deps(
    deps: Dependencies | None, hash_based: bool
) -> tuple[DependencyTuple, ...]:
    ret = []
    if deps is None:
        return ()
    for source_info in deps.static:
        ret.append(
            (
                source_info.name,
                get_dependency_data(source_info, hash_based),
            )
        )
    for modname in deps.nonstatic:
        ret.append((modname, b""))
    return tuple(ret)


def get_dependency_data(source_info: SourceInfo, hash_based: bool) -> bytes:
    if hash_based:
        assert source_info.source is not None
        return importlib.util.source_hash(source_info.source)
    else:
        # pyre-ignore[16]: typeshed doesn't know about this
        return _pack_uint32(source_info.mtime) + _pack_uint32(source_info.size)


def code_to_strict_timestamp_pyc(
    code: CodeType,
    strict_or_static: bool,
    deps: Dependencies | None,
    mtime: int = 0,
    source_size: int = 0,
) -> bytearray:
    "Produce the data for a strict timestamp-based pyc."
    data = bytearray(
        _MAGIC_STRICT_OR_STATIC
        if strict_or_static
        else _MAGIC_NEITHER_STRICT_NOR_STATIC
    )
    data.extend(MAGIC_NUMBER)
    # pyre-ignore[16]: typeshed doesn't know about this
    data.extend(_pack_uint32(0))
    # pyre-ignore[16]: typeshed doesn't know about this
    data.extend(_pack_uint32(mtime))
    # pyre-ignore[16]: typeshed doesn't know about this
    data.extend(_pack_uint32(source_size))
    data.extend(marshal.dumps(get_deps(deps, hash_based=False)))
    data.extend(marshal.dumps(code))
    return data


def code_to_strict_hash_pyc(
    code: CodeType,
    strict_or_static: bool,
    deps: Dependencies | None,
    source_hash: bytes,
    checked: bool = True,
) -> bytearray:
    "Produce the data for a strict hash-based pyc."
    data = bytearray(
        _MAGIC_STRICT_OR_STATIC
        if strict_or_static
        else _MAGIC_NEITHER_STRICT_NOR_STATIC
    )
    data.extend(MAGIC_NUMBER)
    flags = 0b1 | checked << 1
    # pyre-ignore[16]: typeshed doesn't know about this
    data.extend(_pack_uint32(flags))
    assert len(source_hash) == 8
    data.extend(source_hash)
    data.extend(marshal.dumps(get_deps(deps, hash_based=True)))
    data.extend(marshal.dumps(code))
    return data


class StrictSourceFileLoader(SourceFileLoader):
    compiler: Compiler | None = None
    module: ModuleType | None = None

    def __init__(
        self,
        fullname: str,
        path: str,
        import_path: Iterable[str] | None = None,
        stub_path: str | None = None,
        allow_list_prefix: Iterable[str] | None = None,
        allow_list_exact: Iterable[str] | None = None,
        enable_patching: bool = False,
        log_source_load: Callable[[str, str | None, bool], None] | None = None,
        init_cached_properties: None
        | (
            Callable[
                [Mapping[str, str | tuple[str, bool]]],
                Callable[[type[object]], type[object]],
            ]
        ) = None,
        log_time_func: Callable[[], TIMING_LOGGER_TYPE] | None = None,
        use_py_compiler: bool = False,
        # The regexes are parsed on the C++ side, so re.Pattern is not accepted.
        allow_list_regex: Iterable[str] | None = None,
    ) -> None:
        self.name = fullname
        self.path = path
        self.import_path: Iterable[str] = import_path or list(sys.path)
        configured_stub_path = sys._xoptions.get("strict-module-stubs-path") or getenv(
            "PYTHONSTRICTMODULESTUBSPATH"
        )
        if stub_path is None:
            stub_path = configured_stub_path or DEFAULT_STUB_PATH
        if stub_path and not isdir(stub_path):
            raise ValueError(f"Strict module stubs path does not exist: {stub_path}")
        self.stub_path: str = stub_path
        self.allow_list_prefix: Iterable[str] = allow_list_prefix or []
        self.allow_list_exact: Iterable[str] = allow_list_exact or []
        self.allow_list_regex: Iterable[str] = allow_list_regex or []
        self.enable_patching = enable_patching
        self.log_source_load: Callable[[str, str | None, bool], None] | None = (
            log_source_load
        )
        self.bytecode_found = False
        self.bytecode_path: str | None = None
        self.init_cached_properties = init_cached_properties
        self.log_time_func = log_time_func
        self.use_py_compiler = use_py_compiler
        self.strict_or_static: bool = False
        self.is_static: bool = False

    @classmethod
    def ensure_compiler(
        cls,
        path: Iterable[str],
        stub_path: str,
        allow_list_prefix: Iterable[str],
        allow_list_exact: Iterable[str],
        log_time_func: Callable[[], TIMING_LOGGER_TYPE] | None,
        enable_patching: bool = False,
        allow_list_regex: Iterable[str] | None = None,
    ) -> Compiler:
        if (comp := cls.compiler) is None:
            comp = cls.compiler = Compiler(
                path,
                stub_path,
                allow_list_prefix,
                allow_list_exact,
                raise_on_error=True,
                log_time_func=log_time_func,
                enable_patching=enable_patching,
                allow_list_regex=allow_list_regex or [],
            )
        return comp

    def get_compiler(self) -> Compiler:
        return self.ensure_compiler(
            self.import_path,
            self.stub_path,
            self.allow_list_prefix,
            self.allow_list_exact,
            self.log_time_func,
            self.enable_patching,
            self.allow_list_regex,
        )

    def get_code(self, fullname: str) -> CodeType:
        source_path = self.get_filename(fullname)
        source_mtime = None
        source_bytes = None
        source_hash = None
        hash_based = False
        check_source = True
        try:
            bytecode_path = cache_from_source(source_path)
        except NotImplementedError:
            bytecode_path = None
        else:
            bytecode_path = self.bytecode_path = add_strict_tag(
                bytecode_path, self.enable_patching
            )
            try:
                st = self.path_stats(source_path)
            except OSError:
                pass
            else:
                source_mtime = int(st["mtime"])
                try:
                    data = self.get_data(bytecode_path)
                except OSError:
                    pass
                else:
                    self.bytecode_found = True
                    exc_details = {
                        "name": fullname,
                        "path": bytecode_path,
                    }
                    try:
                        flags, strict_or_static = classify_strict_pyc(
                            data, fullname, exc_details
                        )
                        self.strict_or_static = strict_or_static
                        # unmarshal dependencies
                        bytes_data = BytesIO(data)
                        bytes_data.seek(20)
                        deps = marshal.load(bytes_data)
                        hash_based = flags & 0b1 != 0
                        if hash_based:
                            check_source = flags & 0b10 != 0
                            if _imp.check_hash_based_pycs != "never" and (
                                check_source or _imp.check_hash_based_pycs == "always"
                            ):
                                source_bytes = self.get_data(source_path)
                                source_hash = importlib.util.source_hash(source_bytes)
                                # pyre-ignore[16]: typeshed doesn't know about this
                                _validate_hash_pyc(
                                    data[_MAGIC_LEN:],
                                    source_hash,
                                    fullname,
                                    exc_details,
                                )
                                if deps:
                                    validate_dependencies(
                                        self.get_compiler(), deps, hash_based=True
                                    )
                        else:
                            # pyre-ignore[16]: typeshed doesn't know about this
                            _validate_timestamp_pyc(
                                data[_MAGIC_LEN:],
                                source_mtime,
                                st["size"],
                                fullname,
                                exc_details,
                            )
                            if deps:
                                validate_dependencies(
                                    self.get_compiler(), deps, hash_based=False
                                )
                    except (ImportError, EOFError):
                        pass
                    else:
                        # pyre-ignore[16]: typeshed doesn't know about this
                        _bootstrap._verbose_message(
                            "{} matches {}", bytecode_path, source_path
                        )
                        # pyre-ignore[16]: typeshed doesn't know about this
                        return _compile_bytecode(
                            memoryview(data)[bytes_data.tell() :],
                            name=fullname,
                            bytecode_path=bytecode_path,
                            source_path=source_path,
                        )
        if source_bytes is None:
            source_bytes = self.get_data(source_path)
        code_object = self.source_to_code(source_bytes, source_path)
        # pyre-ignore[16]: typeshed doesn't know about this
        _bootstrap._verbose_message("code object from {}", source_path)
        if (
            not sys.dont_write_bytecode
            and bytecode_path is not None
            and source_mtime is not None
        ):
            if self.is_static:
                deps = self.get_compiler().get_dependencies(fullname)
            else:
                deps = None
            if hash_based:
                if source_hash is None:
                    source_hash = importlib.util.source_hash(source_bytes)
                data = code_to_strict_hash_pyc(
                    code_object,
                    self.strict_or_static,
                    deps,
                    source_hash,
                    check_source,
                )
            else:
                data = code_to_strict_timestamp_pyc(
                    code_object,
                    self.strict_or_static,
                    deps,
                    source_mtime,
                    len(source_bytes),
                )
            try:
                # pyre-ignore[16]: typeshed doesn't know about this
                self._cache_bytecode(source_path, bytecode_path, data)
            except NotImplementedError:
                pass
        return code_object

    def should_force_strict(self) -> bool:
        return False

    # pyre-fixme[14]: `source_to_code` overrides method defined in `InspectLoader`
    #  inconsistently.
    def source_to_code(
        self, data: bytes | str, path: str, *, _optimize: int = -1
    ) -> CodeType:
        log_source_load = self.log_source_load
        if log_source_load is not None:
            log_source_load(path, self.bytecode_path, self.bytecode_found)
        # pyre-ignore[28]: typeshed doesn't know about _optimize arg
        code = super().source_to_code(data, path, _optimize=_optimize)
        force = self.should_force_strict()
        if force or "__strict__" in code.co_names or "__static__" in code.co_names:
            # Since a namespace package will never call `source_to_code` (there
            # is no source!), there are only two possibilities here: non-package
            # (submodule_search_paths should be None) or regular package
            # (submodule_search_paths should have one entry, the directory
            # containing the "__init__.py").
            submodule_search_locations = None
            if path.endswith("__init__.py"):
                submodule_search_locations = [path[:12]]
            # Usually _optimize will be -1 (which means "default to the value
            # of sys.flags.optimize"). But this default happens very deep in
            # Python's compiler (in PyAST_CompileObject), so if we just pass
            # around -1 and rely on that, it means we can't make any of our own
            # decisions based on that flag. So instead we do the default right
            # here, so we have the correct optimize flag value throughout our
            # compiler.
            opt = sys.flags.optimize if _optimize == -1 else _optimize
            # Let the ast transform attempt to validate the strict module.  This
            # will return an unmodified module if import __strict__ isn't
            # actually at the top-level
            (
                code,
                is_valid_strict,
                is_static,
            ) = self.get_compiler().load_compiled_module_from_source(
                data,
                path,
                self.name,
                opt,
                submodule_search_locations,
                override_flags=Flags(is_strict=force),
            )
            self.strict_or_static = is_valid_strict or is_static
            self.is_static = is_static
            assert code is not None
            return code

        self.strict_or_static = False
        return code

    def exec_module(self, module: ModuleType) -> None:
        # This ends up being slightly convoluted, because create_module
        # gets called, then source_to_code gets called, so we don't know if
        # we have a strict module until after we were requested to create it.
        # So we'll run the module code we get back in the module that was
        # initially published in sys.modules, check and see if it's a strict
        # module, and then run the strict module body after replacing the
        # entry in sys.modules with a StrictModule entry.  This shouldn't
        # really be observable because no user code runs between publishing
        # the normal module in sys.modules and replacing it with the
        # StrictModule.
        code = self.get_code(module.__name__)
        if code is None:
            raise ImportError(
                f"Cannot import module {module.__name__}; get_code() returned None"
            )
        # fix up the pyc path
        cached = getattr(module.__spec__, "cached", None)
        if cached:
            cached = add_strict_tag(cached, self.enable_patching)
            if module.__spec__ is not None:
                module.__spec__.cached = cached
            if sys.version_info < (3, 15):
                module.__cached__ = cached
        spec: ModuleSpec | None = module.__spec__

        if self.strict_or_static:
            if spec is None:
                raise ImportError(f"Missing module spec for {module.__name__}")

            new_dict = {
                "<fixed-modules>": cast(object, FIXED_MODULES),
                "<builtins>": builtins.__dict__,
                "<init-cached-properties>": self.init_cached_properties,
            }
            if code.co_flags & CI_CO_STATICALLY_COMPILED:
                init_static_python()
                new_dict["<imported-from>"] = code.co_consts[-1]

            new_dict.update(module.__dict__)
            strict_mod = StrictModule(new_dict, self.enable_patching)

            sys.modules[module.__name__] = strict_mod

            exec(code, new_dict)
        else:
            exec(code, module.__dict__)


class StrictSourceFileLoaderWithPatching(StrictSourceFileLoader):
    def __init__(
        self,
        fullname: str,
        path: str,
        import_path: Iterable[str] | None = None,
        stub_path: str | None = None,
        allow_list_prefix: Iterable[str] | None = None,
        allow_list_exact: Iterable[str] | None = None,
        enable_patching: bool = True,
        log_source_load: Callable[[str, str | None, bool], None] | None = None,
        init_cached_properties: None
        | (
            Callable[
                [Mapping[str, str | tuple[str, bool]]],
                Callable[[type[object]], type[object]],
            ]
        ) = None,
        log_time_func: Callable[[], TIMING_LOGGER_TYPE] | None = None,
        use_py_compiler: bool = False,
        # The regexes are parsed on the C++ side, so re.Pattern is not accepted.
        allow_list_regex: Iterable[str] | None = None,
    ) -> None:
        super().__init__(
            fullname,
            path,
            import_path,
            stub_path,
            allow_list_prefix,
            allow_list_exact,
            enable_patching,
            log_source_load,
            init_cached_properties,
            log_time_func,
            use_py_compiler,
            allow_list_regex,
        )


def add_strict_tag(path: str, enable_patching: bool) -> str:
    base, __, ext = path.rpartition(".")
    enable_patching_marker = ".patch" if enable_patching else ""

    return f"{base}.strict{enable_patching_marker}.{ext}"


def _get_supported_file_loaders(
    enable_patching: bool = False,
) -> list[tuple[type[Loader], list[str]]]:
    """Returns a list of file-based module loaders.

    Each item is a tuple (loader, suffixes).
    """
    extensions = ExtensionFileLoader, EXTENSION_SUFFIXES
    source = (
        (
            StrictSourceFileLoaderWithPatching
            if enable_patching
            else StrictSourceFileLoader
        ),
        SOURCE_SUFFIXES,
    )
    bytecode = SourcelessFileLoader, BYTECODE_SUFFIXES
    return [extensions, source, bytecode]


def strict_compile(
    file: str,
    cfile: str,
    dfile: str | None = None,
    doraise: bool = False,
    optimize: int = -1,
    invalidation_mode: PycInvalidationMode | None = None,
    loader_override: object = None,
    loader_options: dict[str, str | int | bool] | None = None,
) -> str | None:
    """Byte-compile one Python source file to Python bytecode, using strict loader.

    :param file: The source file name.
    :param cfile: The target byte compiled file name.
    :param dfile: Purported file name, i.e. the file name that shows up in
        error messages.  Defaults to the source file name.
    :param doraise: Flag indicating whether or not an exception should be
        raised when a compile error is found.  If an exception occurs and this
        flag is set to False, a string indicating the nature of the exception
        will be printed, and the function will return to the caller. If an
        exception occurs and this flag is set to True, a PyCompileError
        exception will be raised.
    :param optimize: The optimization level for the compiler.  Valid values
        are -1, 0, 1 and 2.  A value of -1 means to use the optimization
        level of the current interpreter, as given by -O command line options.
    :return: Path to the resulting byte compiled file.

    Copied and modified from https://github.com/python/cpython/blob/3.6/Lib/py_compile.py#L65

    This version does not support cfile=None, since compileall never passes that.

    """
    modname = file
    for dir in sys.path:
        if file.startswith(dir):
            modname = file[len(dir) :]
            break

    modname = modname.replace("/", ".")
    if modname.endswith("__init__.py"):
        modname = modname[: -len("__init__.py")]
    elif modname.endswith(".py"):
        modname = modname[: -len(".py")]
    modname = modname.strip(".")

    if loader_options is None:
        loader_options = {}

    # TODO we ignore loader_override
    loader = StrictSourceFileLoader(
        modname,
        file,
        import_path=sys.path,
        **loader_options,
    )
    cfile = add_strict_tag(cfile, enable_patching=loader.enable_patching)
    source_bytes = loader.get_data(file)
    try:
        code = loader.source_to_code(source_bytes, dfile or file, _optimize=optimize)
        deps = loader.get_compiler().get_dependencies(modname)
    except Exception as err:
        raise
        py_exc = PyCompileError(err.__class__, err, dfile or file)
        if doraise:
            raise py_exc
        else:
            sys.stderr.write(py_exc.msg + "\n")
            return

    makedirs(dirname(cfile), exist_ok=True)

    if invalidation_mode is None:
        invalidation_mode = _get_default_invalidation_mode()
    if invalidation_mode == PycInvalidationMode.TIMESTAMP:
        source_stats = loader.path_stats(file)
        bytecode = code_to_strict_timestamp_pyc(
            code,
            loader.strict_or_static,
            deps,
            source_stats["mtime"],
            source_stats["size"],
        )
    else:
        source_hash = importlib.util.source_hash(source_bytes)
        bytecode = code_to_strict_hash_pyc(
            code,
            loader.strict_or_static,
            deps,
            source_hash,
            (invalidation_mode == PycInvalidationMode.CHECKED_HASH),
        )

    # pyre-ignore[16]: typeshed doesn't know about this
    loader._cache_bytecode(file, cfile, bytecode)
    return cfile


def init_static_python() -> None:
    """Idempotent global initialization of Static Python.

    Should be called at least once if any Static modules/functions exist.
    """
    watch_sys_modules()
    install_sp_audit_hook()


def install(enable_patching: bool = False) -> None:
    """Installs a loader which is capable of loading and validating strict modules"""
    supported_loaders = _get_supported_file_loaders(enable_patching)

    for index, hook in enumerate(sys.path_hooks):
        if not isinstance(hook, type):
            sys.path_hooks.insert(index, FileFinder.path_hook(*supported_loaders))
            break
    else:
        sys.path_hooks.insert(0, FileFinder.path_hook(*supported_loaders))

    # We need to clear the path_importer_cache so that our new FileFinder will
    # start being used for existing directories we've loaded modules from.
    sys.path_importer_cache.clear()


if __name__ == "__main__":
    install()
    del sys.argv[0]
    mod: object = __import__(sys.argv[0])
    if not isinstance(mod, StrictModule):
        raise TypeError(
            "compiler.strict.loader should be used to run strict modules: "
            + type(mod).__name__
        )
    mod.__main__()
